/*
The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is JSIRC Library

The Initial Developer of the Original Code is New Dimensions Consulting, Inc.

Portions created by the Initial Developer are Copyright (C) 1999
the Initial Developer. All Rights Reserved.

Contributor(s):
  Robert Ginda, rginda@ndcico.com, original author
  Peter Van der Beken, peter.vanderbeken@pandora.be, necko-only version
  Stephen Clavering <mozilla@clav.me.uk>, extensively rewritten for MSNMessenger
*/

// Represents a connection to a Dispatch Server, Notification Server,
// Switchboard, or contact (during file transfer).
// High-level interface for use by ClientConnection and MSNSession is:
//     .connect(aHost, aHost)
//     .disconnect()
//     .sendMsg(aMsg, aBody)
//       aMsg: an array of strings, which will be joined with " " chars
//       aBody: an optional "payload" for the message. If present, its length
//         will be appended to the msg array
//     .isConnected [boolean field]
//     (Also nsIStreamListener, and thus nsIRequestObserver, for internal use.)
// Users of this object should implement:
//     onProtocolMessageReceived(aMsg, aPayload)
//       aMsg: an array of string tokens (which were space-separated)
//       aPayload: a string blob, or null (depending on message type)
// And optionally:
//    onConnectionClosed(aStatus)
//       aStatus: status code for closed connection
//    onConnectionTimedOut()

function BaseConnection(aObserver) {
  this.isConnected = false;
  this.binaryMode = false;
  this.observer = aObserver;

  // Commands which *when received by us* have a payload: a multi-line blob of
  // data whose length is given by the last element of the command. This list
  // must not include commands like QRY, which have a payload when sent by us,
  // but are acknowledged with just a TrID.
  this._payloadCommands = { "NOT": true, "MSG": true };

  this._data = ""; // incoming data buffer. character encoding is unknown

  this._onStartRequestCalled = false;  // helper for onStopRequest

  this._timeout = null; // return value of setTimeout

  // Synchronously open a connection.
  this.connect = function(aHost, aPort) {
    this._log("<o> Connecting to: " + aHost + ":" + aPort);
    this.host = aHost.toLowerCase();
    this.port = aPort;
    this._data = "";
    this.isConnected = true;
    this._onStartRequestCalled = false;

    const sts = Components.classes["@mozilla.org/network/socket-transport-service;1"]
                          .getService(Components.interfaces.nsISocketTransportService);
    this.transport = sts.createTransport(null, 0, aHost, aPort, null);
    // no limit on the output stream buffer
    this._outputStream = this.transport.openOutputStream(0, 4096, -1);
    if (!this._outputStream) {
      throw "Error getting output stream.";
    }
    this._inputStream = this.transport.openInputStream(0, 0, 0);
    if (!this._inputStream) {
      throw "Error getting input stream.";
    }
    const bis = Components.classes["@mozilla.org/binaryinputstream;1"]
                          .createInstance(Components.interfaces.nsIBinaryInputStream);
    bis.setInputStream(this._inputStream);
    this._binaryInputStream = bis;
    const pump = Components.classes["@mozilla.org/network/input-stream-pump;1"]
                           .createInstance(Components.interfaces.nsIInputStreamPump);
    pump.init(this._inputStream, -1, -1, 0, 0, false);
    pump.asyncRead(this, this);
  };

  this.listen = function(aPort) {
    this._log("<o> Listening on aPort " + aPort);
    const sSC = Components.classes["@mozilla.org/network/server-socket;1"];
    if (!sSC) {
      throw("Could not get server socket class.");
    }
    const serverSock = sSC.createInstance();
    if (!serverSock) {
      throw("Could not get server socket.");
    }
    this.serverSock = serverSock.QueryInterface(Components.interfaces.nsIServerSocket);
    this.serverSock.init(aPort, false, -1);
    this.serverSock.asyncListen(this);
  };

  this.onSocketAccepted = function(aSocket, aTransport) {
    this._log("<o> onSocketAccepted");
    this.transport = aTransport;
    this.host = this.transport.host.toLowerCase();
    this.port = this.transport.port;
    this._data = "";
    this.isConnected = true;

    // no limit on the output stream buffer
    this._outputStream = this.transport.openOutputStream(0, 4096, -1);
    if (!this._outputStream) {
      throw "Error getting output stream.";
    }
    this._inputStream = this.transport.openInputStream(0, 0, 0);
    if (!this._inputStream) {
      throw "Error getting input stream.";
    }
    const bis = Components.classes["@mozilla.org/binaryinputstream;1"]
                          .createInstance(Components.interfaces.nsIBinaryInputStream);
    bis.setInputStream(this._inputStream);
    this._binaryInputStream = bis;
    const pump = Components.classes["@mozilla.org/network/input-stream-pump;1"]
                           .createInstance(Components.interfaces.nsIInputStreamPump);
    pump.init(this._inputStream, -1, -1, 0, 0, false);
    pump.asyncRead(this, this);
    this.observer.onConnectionHeard();
    this.stopListening();
  };

  this.disconnect = function() {
    this._log(">o< Disconnect");
    if (this._inputStream) {
      this._inputStream.close();
    }
    if (this._outputStream) {
      this._outputStream.close();
    }
    this.isConnected = false;
  };

  this.stopListening = function() {
    if (this.serverSock) {
      this._log(">o< Stop listening");
      this.serverSock.close();
    }
  };

  onConnectionTimedOut = function() {};

  // Convenience function for sending protocol-level messages in the usual
  // MSN format.
  this.sendMsg = function(aItems, aPayload) {
    if (aPayload) {
      aItems.push(aPayload.length);
    }
    this.sendData(aItems.join(" ") + "\r\n" + (aPayload || ''));
  };

  // Low-level
  this.sendData = function(aString, aTimeout) {
    this._logMsg('>>> ', '.>. ', aString);
    try {
      this._outputStream.write(aString, aString.length);
      if (aTimeout) {
        if (this._timeout) {
          clearTimeout(this._timeout);
        }
        this._timeout = setTimeout(onTimeOutHelper, aTimeout, this);
        // i.e. onTimeOutHelper(this)
      }
    }
    catch(e) {
      this.isConnected = false;
    }
  };

  // nsIStreamListener::onDataAvailable, called by Mozilla's networking code.
  // Buffers the data, and parses it into discrete messages of the form used in
  // MSNP (i.e. space-separated tokens, terminated with \r\n, and possibly a
  // "payload" blob of data following that.
  this.onDataAvailable = function(aRequest, aContext, aString, aSourceOffset, aCount) {
    this._data += this._binaryInputStream.readBytes(aCount);
    if (this._timeout) {
      clearTimeout(this._timeout);
    }

    if (!this.binaryMode) {
      while (this._data) {
        var lineEnd = this._data.indexOf("\r\n");
        var toConsume = lineEnd + 2; // 2 == "\r\n".length
        if (lineEnd == -1) {
          return; // need to buffer more
        }
        var msgdata = this._data.substr(0, lineEnd);
        var msgbits = msgdata.split(" ");
        if (msgbits[0] in this._payloadCommands) {
          var payloadLength = parseInt(msgbits[msgbits.length - 1]);
          // bail out if message isn't entirely here yet
          if (this._data.length < payloadLength + toConsume) {
            return;
          }
          this._dispatchProtocolMessage(msgbits, toConsume, payloadLength);
        }
        else {
          this._dispatchProtocolMessage(msgbits, toConsume, 0);
        }
      }
    }
    else {
      while (this._data) {
        var secondByte = this._data.charCodeAt(1);
        var thirdByte = this._data.charCodeAt(2);
        var bodyLength = secondByte + (thirdByte * 256);
        if (this._data.length < bodyLength + 3) {
          return;
        }
        var bodyBytes = this._data.substr(3, bodyLength);
        this._dispatchBinaryData(bodyBytes, bodyLength);
      }
    }
  };

  this._dispatchProtocolMessage = function(aMsgBits, anMsgBytes, aPayloadLength) {
    this._logMsg('<<< ', '.<. ', this._data.slice(0, anMsgBytes + aPayloadLength));
    const payload = this._data.substr(anMsgBytes, aPayloadLength);
    this._data = this._data.slice(anMsgBytes + aPayloadLength);
    this.observer.onProtocolMessageReceived(aMsgBits, payload);
  };

  this._dispatchBinaryData = function(aBodyBytes, aBodyLength) {
    this._data = this._data.slice(aBodyLength + 3);
    this.observer.onBinaryDataReceived(aBodyBytes, aBodyLength);
  };

  this._logMsg = function(aPrefix, aInfix, aText) {
    var infix = "\n" + aInfix;
    this._log(aPrefix + aText.replace(/\s+$/, '').replace(/\r?\n/g, infix));
  };

  this._log = function(aString) {
    dump(aString + "\n");
  };

  // nsIRequestObserver methods (required by nsIStreamListener)
  this.onStartRequest = function(aRequest, aContext) {
    this._log("BaseConnection.onStartRequest");
    this._onStartRequestCalled = true;
  };

  this.onStopRequest = function(aRequest, aContext, aStatus) {
    this._log("BaseConnection.onStopRequest (" + aStatus + ")");
    if (this._onStartRequestCalled) {
      this.isConnected = false;
    }
    this.observer.onConnectionClosed(aStatus);
  };

  // required by nsISocketListener
  this.onStopListening = function(aSocket, aStatus) {
    this._log("BaseConnection.onStopListening");
    delete this.serverSock;
  };
}


// Wrapper to make "this" be the right object inside .onConnectionTimedOut()
function onTimeOutHelper(aConnection) {
  aConnection.onConnectionTimedOut();
}